from logging import getLogger
from nekbot.protocols import GroupChat
from nekbot.protocols.base.group_chat import GroupChats
from nekbot.protocols.irc.user import UserIRC, UsersIRC
from nekbot.protocols.irc.utils import remove_sharp

__author__ = 'nekmo'

logger = getLogger('nekbot.protocols.irc.group_chat')

class GroupChatIRC(GroupChat):
    def __init__(self, server, channel, name):
        self.channel = channel
        self.server = server
        self.users = UsersIRC(self.server)
        GroupChat.__init__(self, server.protocol, remove_sharp(name))

    def get_users(self, override=True):
        users = UsersIRC(self.server)
        for user in self.channel.users():
            user = UserIRC(self.server, user)
            users[str(user)] = user
        if override: self.users = users
        logger.debug('Users in %s: %s' % (self.name, ', '.join(users)))
        return users

    @property
    def bot(self):
        return UserIRC(self.server, self.server.connection.get_nickname())

    def send_message(self, body):
        self.server.connection.privmsg('#' + self.name,
                                       self.protocol.prepare_message(body))


class GroupChatsIRC(GroupChats):
    pass


# coding=utf-8
from collections import defaultdict
from logging import getLogger

import irc.bot
import irc.strings
from irc.client import ip_numstr_to_quad, NickMask
import threading

from nekbot import settings
from nekbot.core.commands.temp import TempRegex, TempTimeout, CancelTemp
from nekbot.protocols import Protocol
from nekbot.protocols.irc.group_chat import GroupChatsIRC, GroupChatIRC
from nekbot.protocols.irc.message import MessageIRC
from nekbot.protocols.irc.utils import add_sharp, remove_sharp
from nekbot.utils.auth import AuthAddress


"""
"""

__author__ = 'nekmo'

logger = getLogger('nekbot.protocols.irc')


class ServerBot(irc.bot.SingleServerIRCBot):
    def __init__(self, protocol, groupchats_list, username, realname, server, port=6667):
        self.identified_command = None
        self.users_by_username = {} # Usuarios en el servidor conocidos.
        self.groupchats = {}  # {'groupchat': <Groupchat object>}
        self.groupchats_list = groupchats_list # ['groupchat1', 'groupchat2']
        self.protocol = protocol
        self.domain = server
        irc.bot.SingleServerIRCBot.__init__(self, [(server, port)], username, realname)

    def input_message(self, event):
        logger.debug('Incoming message: %s' % event.arguments[0])
        self.protocol.propagate('message', MessageIRC(self, event))

    def _execute_identified_command(self, username, command=None, result=None):
        if command is None:
            command = self.identified_command
        if result is None:
            result = settings.IDENTIFIED_COMMANDS[command]
        self.connection.send_raw(command % username)
        pattern = result[0].format(username=username)
        temp = TempRegex(self.protocol, pattern, timeout=6, no_raise=True)
        for msg in temp.read():
            if isinstance(msg, TempTimeout): break
            if msg == CancelTemp: break
            if msg.server != self: continue
            temp.done()
            return msg
        return False

    def _get_identified_command(self):
        for command, result in settings.IDENTIFIED_COMMANDS.items():
            if self._execute_identified_command(self.connection.get_nickname(), command, result) is not False:
                logger.debug('Identified Command for %s: %s' % (self.domain, command))
                self.identified_command = command
                return result
        logger.info('Identified Command unknown for server %s' % self.domain)
        self.identified_command = False
        return False

    def get_identified_command(self):
        l  = threading.Thread(target=self._get_identified_command)
        l.start()

    def get_identified(self, username):
        """IRC no tiene una forma sencilla de comprobar si un usuario está autenticado.
        Intentar descubrirlo.
        """
        if self.identified_command is False:
            # No hay un comando conocido para saber si el usuario está conectado
            return False
        if self.identified_command is None:
            return False
            # self.get_identified_command()
        msg = self._execute_identified_command(username)
        if msg is False: return False
        result = settings.IDENTIFIED_COMMANDS[self.identified_command]
        if msg.match[0] == result[1]:
            return True
        return False

    def on_welcome(self, connection, event):
        self.get_identified_command()
        for groupchat in self.groupchats_list:
            self.join_groupchat(add_sharp(groupchat))

    def join_groupchat(self, channel, key=''):
        self.connection.join(channel, key)

    def on_privmsg(self, connection, event):
        self.input_message(event)

    def on_pubmsg(self, connection, event):
        self.input_message(event)

    def on_privnotice(self, channel, event):
        self.input_message(event)

    def on_pubnotice(self, channel, event):
        self.input_message(event)

    def on_dccmsg(self, channel, event):
        # non-chat DCC messages are raw bytes; decode as text
        text = event.arguments[0].decode('utf-8')
        channel.privmsg("You said: " + text)

    def on_join(self, connection, event):
        if event.target not in self.groupchats:
            # Se está entrando a la sala por primera vez
            groupchat = GroupChatIRC(self, self.channels[event.target], event.target)
            self.groupchats[str(groupchat)] = groupchat
            self.protocol.groupchats[str(groupchat)] = groupchat
            groupchat.get_users()
        else:
            # Un usuario ha entrado a al sala
            user = NickMask(event.source)
            self.groupchats[remove_sharp(event.target)].users[str(user)] = user

    def on_part(self, connection, event):
        # {'source': u'nekmo!~nekmo@localhost.localdomain',
        # 'type': u'part', 'target': u'#testing', 'arguments': [u'Saliendo']}
        user = NickMask(event.source)
        del self.groupchats[remove_sharp(event.target)].users[str(user)]

    def on_mode(self, connection, event):
        pass

    def on_nicknameinuse(self, connection, event):
        connection.nick(connection.get_nickname() + "_")

    def on_dccchat(self, channel, event):
        if len(event.arguments) != 2:
            return
        args = event.arguments[1].split()
        if len(args) == 4:
            try:
                address = ip_numstr_to_quad(args[2])
                port = int(args[3])
            except ValueError:
                return
            self.dcc_connect(address, port)

            # def do_command(self, e, cmd):
            # nick = e.source.nick
            # c = self.connection
            #
            #     if cmd == "disconnect":
            #         self.disconnect()
            #     elif cmd == "die":
            #         self.die()
            #     elif cmd == "stats":
            #         for chname, chobj in self.channels.items():
            #             c.notice(nick, "--- Channel statistics ---")
            #             c.notice(nick, "Channel: " + chname)
            #             users = sorted(chobj.users())
            #             c.notice(nick, "Users: " + ", ".join(users))
            #             opers = sorted(chobj.opers())
            #             c.notice(nick, "Opers: " + ", ".join(opers))
            #             voiced = sorted(chobj.voiced())
            #             c.notice(nick, "Voiced: " + ", ".join(voiced))
            #     elif cmd == "dcc":
            #         dcc = self.dcc_listen()
            #         c.ctcp("DCC", nick, "CHAT chat %s %d" % (
            #             ip_quad_to_numstr(dcc.localaddress),
            #             dcc.localport))
            #     else:
            #         c.notice(nick, "Not understood: " + cmd)


class Irc(Protocol):
    features = ['notice']

    def __init__(self, nekbot):
        self.servers = []
        self.groupchats = GroupChatsIRC(self)
        self._addresses = [AuthAddress(addr) for addr in settings.IRC_GROUPCHATS]  # groupchat@server
        self._groupchats_by_server = defaultdict(list)  # {server: [groupchat@server]}
        super(Irc, self).__init__(nekbot)

    def prepare_message(self, body):
        if not isinstance(body, (str, unicode)):
            body = str(body)
        body = body.decode('utf-8')
        return body

    def init(self):
        for address in self._addresses:
            self._groupchats_by_server[address.endpoint].append(address)
        for server, auth in settings.IRC_AUTHS.items():
            # El identificador es addr.user, pero en realidad es addr.groupchat. Esto es porque es una
            # clase abstraida para obtener user@host, donde user en realidad es el groupchat.
            server = AuthAddress(server)
            serverbot = ServerBot(
                self,  # protocol instance
                [addr.user for addr in self._groupchats_by_server[str(server)]],  # rooms
                auth['username'],  # username
                auth.get('realname', auth['username']),  # Realname
                server.host,  # hostname without port
                server.port if server.port else 6667  # server port
            )
            self.servers.append(serverbot)

    def run(self):
        for server in self.servers:
            server.start()


from irc.client import NickMask
import traceback
import sys
from nekbot.protocols import User
from nekbot.protocols.base.user import Users

__author__ = 'nekmo'


class UserIRC(User):
    def __init__(self, server, user):
        user = NickMask(user)
        # user.nick
        # user.userhost
        # user.host
        # user.user
        self.server = server
        self._id = None
        User.__init__(self, server.protocol, user.nick)

    def send_message(self, body, notice=False):
        if notice:
            self.server.connection.notice(self.username, self.protocol.prepare_message(body))
        else:
            self.server.connection.privmsg(self.username, self.protocol.prepare_message(body))

    @property
    def id(self):
        return self._id

    def get_id(self):
        if self._id:
            return self._id
        if self.username in self.server.users_by_username:
            self._id = self.server.users_by_username[self.username].id
            return self._id
        if not self.server.get_identified(self.username):
            self._id = None
        else:
            self._id = '%s@%s' % (self.username, self.server.domain)
        self.server.users_by_username[self.username] = self
        return self._id

class UsersIRC(Users):
    def __init__(self, server):
        self.server = server
        super(UsersIRC, self).__init__(server.protocol)


from nekbot.protocols import Message
from nekbot.protocols.irc.utils import remove_sharp
from nekbot.protocols.irc.user import UserIRC

__author__ = 'nekmo'


class MessageIRC(Message):
    def __init__(self, server, event):
        self.server = server
        self.event = event
        body = event.arguments[0]
        if self.is_groupchat:
            groupchat = server.groupchats[remove_sharp(event.target)]
        else:
            groupchat = None
        user = UserIRC(server, event.source)
        super(MessageIRC, self).__init__(server.protocol, body, user, groupchat)

    @property
    def is_groupchat(self):
        if self.is_private:
            return False
        elif self.is_public:
            return True
        else:
            raise NotImplementedError('Improperly implementation of is_groupchat.')

    @property
    def is_from_me(self):
        return self.server.connection.get_nickname() == self.user.username

    @property
    def is_private(self):
        return self.event.type in ['privmsg', 'privnotice']

    @property
    def is_public(self):
        return self.event.type in ['pubmsg']

    def _copy(self):
        return self.__class__(self.server, self.event)




IDENTIFIED_COMMANDS = {
    'NS STATUS %s': ['STATUS {username} (\d)', '3'],
    'PRIVMSG NickServ ACC %s': ['{username} ACC (\d)', '3'],
}

IRC_AUTHS = {
    # 'localhost': {'username': 'Botname', 'realname': 'Bot Real Name'},
}

IRC_GROUPCHATS = [
    # 'groupchat@localhost',
]

IRC_PERMISSIONS = {
    # 'nekmo@localhost': ['root'],
}


# stevedore/example/setup.py
from setuptools import setup, find_packages
import re
from pip.req import parse_requirements
import uuid

VERSION = open('VERSION').read().replace('\n', '')
ROOT_INCLUDE = ['requirements.txt', 'VERSION', 'LICENSE.txt']

requirements = parse_requirements('requirements.txt', session=uuid.uuid1())
install_requires = [str(ir.req) for ir in requirements if not ir.url]

setup(
    name='nekbot.protocols.irc',
    namespace_packages = ['nekbot.protocols'],
    version=VERSION,

    description='IRC Protocol for Nekbot, a modular multiprotocol bot.',

    author='Nekmo',
    author_email='contacto@nekmo.com',

    url='https://bitbucket.org/Nekmo/nekbot.protocols.irc',
    
    download_url='https://bitbucket.org/Nekmo/nekbot.protocols.irc/get/default.tar.gz',

    classifiers=[
        'Natural Language :: Spanish',
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Topic :: Communications :: Chat',
        'Topic :: Communications :: Chat :: Internet Relay Chat',
        'Topic :: Communications :: Conferencing',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    platforms=['linux'],

    scripts=[],
    
    provides=['nekbot.protocols.irc',
              ],
    
    install_requires=install_requires,

    packages=['nekbot', 'nekbot.protocols', 'nekbot.protocols.irc'],
    include_package_data=True,

    keywords=['nekbot', 'bot', 'irc', 'chat'],

    entry_points={
        'nekbot.protocols': [
            'irc = nekbot.protocols.irc:MyIRC',
        ],
    },
        
    package_data = {'': ROOT_INCLUDE},

    zip_safe=False,
)